from typing import Literal
import uuid

import pydantic


class BaseSNSPayload(pydantic.BaseModel):
    MessageId: uuid.UUID
    TopicArn: str
    Message: str
    # This must be left as a string and not parsed into datetime so that
    # signature calculation still works
    Timestamp: str
    SignatureVersion: str
    Signature: pydantic.Base64Bytes
    SigningCertURL: pydantic.HttpUrl


class SubscriptionConfirmation(BaseSNSPayload):
    Type: Literal["SubscriptionConfirmation"]
    Token: str
    SubscribeURL: pydantic.HttpUrl


class UnsubscribeConfirmation(BaseSNSPayload):
    Type: Literal["UnsubscribeConfirmation"]
    Token: str
    SubscribeURL: pydantic.HttpUrl


class Notification(BaseSNSPayload):
    Type: Literal["Notification"]
    Subject: str | None = None
    UnsubscribeURL: pydantic.HttpUrl


type AnySNSPayload = SubscriptionConfirmation | UnsubscribeConfirmation | Notification


class SNSPayload(pydantic.RootModel[AnySNSPayload]):
    root: AnySNSPayload = pydantic.Field(discriminator="Type")



import logging
import re

from django.conf import settings
from django.http import HttpRequest, HttpResponse, HttpResponseBadRequest
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View
import pydantic_core

from .types import (
    Notification,
    SNSPayload,
    SubscriptionConfirmation,
    UnsubscribeConfirmation,
)
from .utils import confirm_subscription, verify_notification

logger = logging.getLogger(__name__)


@method_decorator(csrf_exempt, name="dispatch")
class SNSEndpoint(View):
    topic_type_header: str = "HTTP_X_AMZ_SNS_TOPIC_ARN"
    cert_domain_settings_key: str = "SNS_CERT_DOMAIN_REGEX"
    sns_verify_settings_key: str = "SNS_VERIFY_CERTIFICATE"
    topic_settings_key: str = ""

    def handle_message(self, message: str, notification: Notification) -> None:
        """
        Process the SNS message.
        """
        raise NotImplementedError

    def post(self, request: HttpRequest) -> HttpResponse:
        """
        Validate and handle an SNS message.
        """
        # Check the topic if specified by a settings key
        topic_allowlist = self.get_topic_allowlist()
        if topic_allowlist is not None:
            if self.topic_type_header not in request.META:
                return HttpResponseBadRequest("No TopicArn Header")

            # Check to see if the topic is in the settings
            if request.META[self.topic_type_header] not in topic_allowlist:
                return HttpResponseBadRequest("Bad Topic")

        # Parse and validate the request body
        try:
            payload = SNSPayload.model_validate_json(request.body).root
        except pydantic_core.ValidationError:
            logger.exception("Invalid payload")
            return HttpResponseBadRequest("Invalid payload")

        # Confirm that the signing certificate is hosted on a correct domain
        # AWS by default uses sns.{region}.amazonaws.com
        pattern = self.get_cert_domain_pattern()
        if not payload.SigningCertURL.host or not re.search(
            pattern, payload.SigningCertURL.host
        ):
            logger.warning(
                "Improper Certificate Location %s",
                payload.SigningCertURL,
            )
            return HttpResponseBadRequest("Improper Certificate Location")

        # Verify that the notification is signed by Amazon
        if self.get_cert_verification_enabled() and not verify_notification(payload):
            logger.error("Cert verification failed")
            return HttpResponseBadRequest("Improper Signature")

        # Handle subscription confirmations
        if isinstance(payload, SubscriptionConfirmation):
            return confirm_subscription(payload)

        # Handle unsubscribe confirmations
        if isinstance(payload, UnsubscribeConfirmation):
            # Don't handle unsubscribe notification here, just remove
            # this endpoint from AWS console. Return 200 status
            # so redelivery of this message doesnt occur.
            logger.info("UnsubscribeConfirmation Not Handled")
            return HttpResponse("UnsubscribeConfirmation Not Handled")

        message = payload.Message
        logger.info(
            "SNS Notification received",
            extra=dict(
                message_type=payload.Type,
                sns_payload=request.body,
                payload_message=payload,
            ),
        )
        self.handle_message(message, payload)
        return HttpResponse("OK")

    def get_topic_allowlist(self) -> list[str] | None:
        return getattr(settings, self.topic_settings_key, None)

    def get_cert_domain_pattern(self) -> str:
        return getattr(
            settings,
            self.cert_domain_settings_key,
            r"sns.[a-z0-9\-]+.amazonaws.com$",
        )

    def get_cert_verification_enabled(self) -> bool:
        return getattr(
            settings,
            self.sns_verify_settings_key,
            True,
        )



import os

from cryptography import x509
from django.conf import settings
from django.test import TestCase, override_settings

from ..types import Notification, SubscriptionConfirmation, UnsubscribeConfirmation
from .test_data.notifications import (
    SNS_NOTIFICATION,
    SNS_NOTIFICATION_NO_SUBJECT,
    SNS_SUBSCRIPTION_NOTIFICATION,
    SNS_UNSUBSCRIBE_NOTIFICATION,
)

DIRNAME, _ = os.path.split(os.path.abspath(__file__))


@override_settings(SNS_STORY_TOPIC_ARN=["arn:aws:sns:us-west-2:123456789012:MyTopic"])
class SNSBaseTest(TestCase):
    old_topics = getattr(settings, "", None)

    sns_notification = Notification.model_validate(SNS_NOTIFICATION)
    sns_notification_no_subject = Notification.model_validate(
        SNS_NOTIFICATION_NO_SUBJECT
    )
    sns_confirmation = SubscriptionConfirmation.model_validate(
        SNS_SUBSCRIPTION_NOTIFICATION
    )
    sns_unsubscribe = UnsubscribeConfirmation.model_validate(
        SNS_UNSUBSCRIBE_NOTIFICATION
    )

    keyfileobj = open(DIRNAME + ("/test_data/example.pem"))
    pemfile = keyfileobj.read().encode()
    x509_cert = x509.load_pem_x509_certificate(pemfile)



SNS_SUBSCRIPTION_NOTIFICATION = {
    "Type": "SubscriptionConfirmation",
    "MessageId": "ad01b87e-626a-443b-9b28-3c40ff5ebea5",
    "Token": "2336412f37fb687f5d51e6e241da92fd744699239e28eb25ebcef294aa79b3e339f26b7cc430ce0871bab7b13d28668befdbbf5597a17db06049c896cd457f1b2f3c582a403affaafb2ae9baa6d15d0afc51c0050f0d419a632d55d7a4a024d0d11953fac4753f008d78ea9c5085ef599c4e64846db77b760c00e4377a012448",
    "TopicArn": "arn:aws:sns:ap-southeast-2:919599206538:test-example",
    "Message": "You have chosen to subscribe to the topic arn:aws:sns:ap-southeast-2:919599206538:test-example.\nTo confirm the subscription, visit the SubscribeURL included in this message.",
    "SubscribeURL": "https://sns.ap-southeast-2.amazonaws.com/?Action=ConfirmSubscription&TopicArn=arn:aws:sns:ap-southeast-2:919599206538:test-example&Token=2336412f37fb687f5d51e6e241da92fd744699239e28eb25ebcef294aa79b3e339f26b7cc430ce0871bab7b13d28668befdbbf5597a17db06049c896cd457f1b2f3c582a403affaafb2ae9baa6d15d0afc51c0050f0d419a632d55d7a4a024d0d11953fac4753f008d78ea9c5085ef599c4e64846db77b760c00e4377a012448",
    "Timestamp": "2018-01-07T04:07:16.301Z",
    "SignatureVersion": "1",
    "Signature": "Ps5DODb4HQyVeZnuqnjMxwAXo1xg2/cPiXJ53quM/LzxUYbeh5tWmtOZr5l/kvgVe1DECaEYS5cSgwTrFjEJBt+M1EVXK4AaeiHyzk8LlghA5dhoICg+wkS3kHiQYKvU/JczGODImx0qA2e2HnAZ3DqPb+sY3EI10xebvnsvHoHJ6YqnJdC6L5vjFj6SW1xb+2oAo/teHeZYm/Fb7z17uGL06e80O43y+o/v4M8lZ1qZsUWZzluP48GciBegw4++AWUxNzEZSZKoV3AzCjObLeUHVI6bwTf9+xpKU9vhPPbLwIJyOUXEKlJQEPqgRoFkB+Ts60q2Ikg7h54QZW5CGQ==",
    "SigningCertURL": "https://sns.ap-southeast-2.amazonaws.com/SimpleNotificationService-433026a4050d206028891664da859041.pem",
}

SNS_UNSUBSCRIBE_NOTIFICATION = {
    "Type": "UnsubscribeConfirmation",
    "MessageId": "ad01b87e-626a-443b-9b28-3c40ff5ebea5",
    "Token": "2336412f37fb687f5d51e6e241da92fd744699239e28eb25ebcef294aa79b3e339f26b7cc430ce0871bab7b13d28668befdbbf5597a17db06049c896cd457f1b2f3c582a403affaafb2ae9baa6d15d0afc51c0050f0d419a632d55d7a4a024d0d11953fac4753f008d78ea9c5085ef599c4e64846db77b760c00e4377a012448",
    "TopicArn": "arn:aws:sns:ap-southeast-2:919599206538:test-example",
    "Message": "You have chosen to subscribe to the topic arn:aws:sns:ap-southeast-2:919599206538:test-example.\nTo confirm the subscription, visit the SubscribeURL included in this message.",
    "SubscribeURL": "https://sns.ap-southeast-2.amazonaws.com/?Action=ConfirmSubscription&TopicArn=arn:aws:sns:ap-southeast-2:919599206538:test-example&Token=2336412f37fb687f5d51e6e241da92fd744699239e28eb25ebcef294aa79b3e339f26b7cc430ce0871bab7b13d28668befdbbf5597a17db06049c896cd457f1b2f3c582a403affaafb2ae9baa6d15d0afc51c0050f0d419a632d55d7a4a024d0d11953fac4753f008d78ea9c5085ef599c4e64846db77b760c00e4377a012448",
    "Timestamp": "2018-01-07T04:07:16.301Z",
    "SignatureVersion": "1",
    "Signature": "Ps5DODb4HQyVeZnuqnjMxwAXo1xg2/cPiXJ53quM/LzxUYbeh5tWmtOZr5l/kvgVe1DECaEYS5cSgwTrFjEJBt+M1EVXK4AaeiHyzk8LlghA5dhoICg+wkS3kHiQYKvU/JczGODImx0qA2e2HnAZ3DqPb+sY3EI10xebvnsvHoHJ6YqnJdC6L5vjFj6SW1xb+2oAo/teHeZYm/Fb7z17uGL06e80O43y+o/v4M8lZ1qZsUWZzluP48GciBegw4++AWUxNzEZSZKoV3AzCjObLeUHVI6bwTf9+xpKU9vhPPbLwIJyOUXEKlJQEPqgRoFkB+Ts60q2Ikg7h54QZW5CGQ==",
    "SigningCertURL": "https://sns.ap-southeast-2.amazonaws.com/SimpleNotificationService-433026a4050d206028891664da859041.pem",
}

SNS_NOTIFICATION = {
    "Type": "Notification",
    "MessageId": "62acea4d-630a-5a7f-8a4e-b886c17258dd",
    "TopicArn": "arn:aws:sns:ap-southeast-2:919599206538:test-example",
    "Subject": "With Subject",
    "Message": '{"hello": "world"}',
    "Timestamp": "2018-01-07T04:21:12.742Z",
    "SignatureVersion": "1",
    "Signature": "V0APJSXwl5ubpJYJUVeV8PfP7LVnVWk7+W/J2KITo2ZIQlS4Z3KQ/o2nTci6Dg+v+IzDGBmFS2SNBZJiikXtTVvJeqRIb3uqKs2aDSH+cm8PphJhkhfoxNTcwD6J1n3uEgVz7n7ZdHphQGfGuKVT9d8IbPeczBKZ8twB7HLlr/v5NhMpRhYqYMypm916rO5hgyWaDqO3w5o3mzYLevYW4KuhMvdwLOQxslF+SCUSbQGqAbgm5jY1XZxN5wv0bF/3BgXJirmlMOTmk0O+xkTj4rhs2u9vlfMV5thRXqCo7LUSIGGMuXvBjxAF9Ph/QcKTFIHlitBgjWKfdoUWeblNkg==",
    "SigningCertURL": "https://sns.ap-southeast-2.amazonaws.com/SimpleNotificationService-433026a4050d206028891664da859041.pem",
    "UnsubscribeURL": "https://sns.ap-southeast-2.amazonaws.com/?Action=Unsubscribe&SubscriptionArn=arn:aws:sns:ap-southeast-2:919599206538:test-example:358ec852-8d21-4777-8175-0433ad263cb3",
    "MessageAttributes": {
        "AWS.SNS.MOBILE.MPNS.Type": {"Type": "String", "Value": "token"},
        "AWS.SNS.MOBILE.MPNS.NotificationClass": {
            "Type": "String",
            "Value": "realtime",
        },
        "AWS.SNS.MOBILE.WNS.Type": {"Type": "String", "Value": "wns/badge"},
    },
}

SNS_NOTIFICATION_NO_SUBJECT = {
    "Type": "Notification",
    "MessageId": "cd2fe5d6-f321-5602-b250-b71d51fd1689",
    "TopicArn": "arn:aws:sns:ap-southeast-2:919599206538:test-example",
    "Message": '{"hello": "world"}',
    "Timestamp": "2018-01-07T04:22:52.993Z",
    "SignatureVersion": "1",
    "Signature": "ivW/VU2bxFcAkvAqx+PAor33My/jsXIxcz9vRvU4k2u2uctXLT4GvNqYhim4ISdRZ8yXKnaHkWa/K3N2awf9KWfpXnshRlvAAnPcIXRrIAXFy49I0V9/grjYJOsKDkKS/jXh23AfhUfLNyltjDqpY1Fy9a5c7OqU7zLYJmjtTQQSOr3uMV7RsEtbx0M1SO3z9TeF7N7bXGxoKZboFx2jxSjSeuMRMvjlX/ut7fMyFbyuiXhkEg1WXJRuBLWqeistVdVPLr1r5ol8bwUiLOkjnG1xSB21H5Y2LCp5PcxPY6Bhg7VSiSNX1FPLFQdQ8s5AViOYir+VReE4KEuB7prduw==",
    "SigningCertURL": "https://sns.ap-southeast-2.amazonaws.com/SimpleNotificationService-433026a4050d206028891664da859041.pem",
    "UnsubscribeURL": "https://sns.ap-southeast-2.amazonaws.com/?Action=Unsubscribe&SubscriptionArn=arn:aws:sns:ap-southeast-2:919599206538:test-example:358ec852-8d21-4777-8175-0433ad263cb3",
    "MessageAttributes": {
        "AWS.SNS.MOBILE.MPNS.Type": {"Type": "String", "Value": "token"},
        "AWS.SNS.MOBILE.MPNS.NotificationClass": {
            "Type": "String",
            "Value": "realtime",
        },
        "AWS.SNS.MOBILE.WNS.Type": {"Type": "String", "Value": "wns/badge"},
    },
}



from unittest.mock import MagicMock, Mock, patch

from cryptography import x509
from django.conf import settings
from requests.exceptions import HTTPError
import pydantic

from ..utils import confirm_subscription, get_x509_cert, verify_notification
from .helpers import SNSBaseTest


class VerificationTest(SNSBaseTest):
    @patch("django_sns_view.utils.requests.get")
    def test_get_x509_cert(self, mock: MagicMock) -> None:
        """Test the get_pemfile util"""
        responsemock = Mock()
        mock.return_value = responsemock
        responsemock.text = self.pemfile
        result = get_x509_cert("http://www.fakeurl.com")

        mock.assert_called_with("http://www.fakeurl.com")
        self.assertIsInstance(result, x509.Certificate)

    @patch("django_sns_view.utils.requests.get")
    def test_bad_keyfile(self, mock: MagicMock) -> None:
        """Test a non-valid keyfile"""
        responsemock = Mock()
        responsemock.read.return_value = "Not A Certificate"
        mock.return_value = responsemock

        with self.assertRaises(ValueError) as context_manager:
            get_x509_cert("http://www.fakeurl.com")

        the_exception = context_manager.exception
        self.assertTrue(the_exception.args[0].startswith("Unable to load PEM file"))

    @patch("django_sns_view.utils.get_x509_cert")
    def test_verify_notification_with_subject(self, mock: MagicMock) -> None:
        """Test the verification of a valid notification with Subject"""
        mock.return_value = self.x509_cert
        result = verify_notification(self.sns_notification)
        self.assertTrue(result)

    @patch("django_sns_view.utils.get_x509_cert")
    def test_verify_notification_no_subject(self, mock: MagicMock) -> None:
        """Test the verification of a valid notification without Subject"""
        mock.return_value = self.x509_cert
        result = verify_notification(self.sns_notification_no_subject)
        self.assertTrue(result)


class ConfirmSubscriptionTest(SNSBaseTest):
    @patch("django_sns_view.utils.requests.get")
    def test_successful_confirm_subscription(self, mock: MagicMock) -> None:
        """Test a successful subscription confirmation"""
        mock_resp = Mock()
        mock_resp.raise_for_status = Mock()
        mock_resp.status_code = 200

        mock.return_value = mock_resp

        response = confirm_subscription(self.sns_confirmation)

        mock.assert_called_with(str(self.sns_confirmation.SubscribeURL))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content.decode("ascii"), "OK")

    @patch("django_sns_view.utils.requests.get")
    def test_fail_confirm_subscription(self, mock: MagicMock) -> None:
        """Test a successful subscription confirmation"""
        mock_resp = Mock()
        mock_resp.raise_for_status = Mock()
        mock_resp.raise_for_status.side_effect = HTTPError("site is down")
        mock_resp.status_code = 503
        mock.return_value = mock_resp

        self.assertRaises(HTTPError, confirm_subscription, self.sns_confirmation)
        mock.assert_called_with(str(self.sns_confirmation.SubscribeURL))

    def test_bad_url(self) -> None:
        """Test to make sure an invalid URL isn't requested"""
        old_setting = getattr(settings, "SNS_SUBSCRIBE_DOMAIN_REGEX", None)
        settings.SNS_SUBSCRIBE_DOMAIN_REGEX = r"sns.[a-z0-9\-]+.amazonaws.com$"
        notification = self.sns_confirmation.model_copy()
        notification.SubscribeURL = pydantic.HttpUrl("http://anon.amazonaws.com")
        result = confirm_subscription(notification)

        self.assertEqual(result.status_code, 400)
        self.assertEqual(result.content.decode("ascii"), "Improper Subscription Domain")

        if old_setting is not None:
            settings.SNS_SUBSCRIBE_DOMAIN_REGEX = old_setting



import os

SECRET_KEY = "foobar"
DEBUG = True
SITE_ID = 1

APP_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.admindocs",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.messages",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
    "django.contrib.sitemaps",
    "django.contrib.sites",
    "django_sns_view",
]

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django.middleware.security.SecurityMiddleware",
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": (
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "django.template.context_processors.i18n",
            )
        },
    }
]



from copy import deepcopy
from unittest.mock import MagicMock, patch
import json

from django.conf import settings
from django.http import HttpResponse, HttpResponseNotAllowed
from django.test import RequestFactory
from django.test.utils import override_settings

from ..types import Notification
from ..views import SNSEndpoint
from .helpers import SNSBaseTest
from .test_data.notifications import SNS_NOTIFICATION


@override_settings(SNS_VERIFY_CERTIFICATE=False)
class SNSEndpointTestCase(SNSBaseTest):
    def setUp(self) -> None:
        self.factory = RequestFactory()
        self.request = self.factory.post("/")
        self.endpoint = SNSEndpoint.as_view()
        self.request.META["HTTP_X_AMZ_SNS_TOPIC_ARN"] = settings.SNS_STORY_TOPIC_ARN[0]  # type:ignore[misc]
        self.request.META["HTTP_X_AMZ_SNS_MESSAGE_TYPE"] = "Notification"

    def test_non_post_httpnotallowed(self) -> None:
        """
        Test that GET requests to the endpoint return a
        HttpResponseNotAllowed
        """
        request = self.factory.get("/")
        response = self.endpoint(request)
        self.assertIsInstance(response, HttpResponseNotAllowed)

    @patch.object(
        SNSEndpoint,
        "handle_message",
        lambda _, message, notification: HttpResponse("OK"),
    )
    def test_success(self) -> None:
        """Test a successful request"""
        self.request._body = self.sns_notification.model_dump_json().encode()
        response = self.endpoint(self.request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.content.decode("ascii"),  # type:ignore[attr-defined]
            "OK",
        )

    @override_settings(
        SNS_TEST_TOPIC_ARN=["arn:aws:sns:us-west-2:123456789012:MyTopic"]
    )
    @patch.object(SNSEndpoint, "topic_settings_key", "SNS_TEST_TOPIC_ARN")
    def test_no_topic_header(self) -> None:
        """Test the results if the request does not have a topic header"""
        request = self.factory.post("/")
        request.META["HTTP_X_AMZ_SNS_MESSAGE_TYPE"] = "Notification"
        request._body = self.sns_notification.model_dump_json().encode()
        response = self.endpoint(request)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.content.decode("ascii"),  # type:ignore[attr-defined]
            "No TopicArn Header",
        )

    @override_settings(SNS_TEST_TOPIC_ARN=["Diddly Doo"])
    @patch.object(SNSEndpoint, "topic_settings_key", "SNS_TEST_TOPIC_ARN")
    def test_bad_topic(self) -> None:
        """Test the response if the topic does not match the settings"""
        self.request._body = self.sns_notification.model_dump_json().encode()
        response = self.endpoint(self.request)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.content.decode("ascii"),  # type:ignore[attr-defined]
            "Bad Topic",
        )

    def test_invalid_notification_json(self) -> None:
        """Test if the notification does not have a JSON body"""
        self.request._body = b"This Is Not JSON"
        response = self.endpoint(self.request)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.content.decode("ascii"),  # type:ignore[attr-defined]
            "Invalid payload",
        )

    def test_bad_certificate_url(self) -> None:
        """Test an unknown certificate hostname"""
        notification = deepcopy(SNS_NOTIFICATION)
        notification["SigningCertURL"] = "https://baddomain.com/cert.pem"
        self.request._body = json.dumps(notification).encode()
        response = self.endpoint(self.request)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.content.decode("ascii"),  # type:ignore[attr-defined]
            "Improper Certificate Location",
        )

    @patch("django_sns_view.views.confirm_subscription")
    def test_confirm_subscription_called(self, mock: MagicMock) -> None:
        """
        Test that confirm_subscription is called when sns
        sends a SubscriptionConfirmation notification
        """
        mock.return_value = "Confirmed"
        self.request.META["HTTP_X_AMZ_SNS_MESSAGE_TYPE"] = "SubscriptionConfirmation"
        self.request._body = self.sns_confirmation.model_dump_json().encode()
        response = self.endpoint(self.request)
        self.assertTrue(mock.called)
        self.assertEqual(response, "Confirmed")

    def test_unsubscribe_confirmation_not_handled(self) -> None:
        """Test that an unsubscribe notification is properly ignored"""
        self.request.META["HTTP_X_AMZ_SNS_MESSAGE_TYPE"] = "UnsubscribeConfirmation"
        self.request._body = self.sns_unsubscribe.model_dump_json().encode()
        response = self.endpoint(self.request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.content.decode("ascii"),  # type:ignore[attr-defined]
            "UnsubscribeConfirmation Not Handled",
        )

    @patch.object(SNSEndpoint, "handle_message")
    def test_handle_message_sucessfully_called(self, mock: MagicMock) -> None:
        """
        Test that the handle message method is called successfully,
        and that it has the right arguments passed.
        """
        mock.return_value = True
        self.request._body = self.sns_notification.model_dump_json().encode()
        response = self.endpoint(self.request)
        mock.assert_called_once_with(
            SNS_NOTIFICATION.get("Message"),
            Notification.model_validate(SNS_NOTIFICATION),
        )
        self.assertEqual(response.status_code, 200)



from functools import lru_cache
import logging
import re

from cryptography import x509
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from django.conf import settings
from django.http import HttpResponse, HttpResponseBadRequest
from django.utils.encoding import smart_bytes
from requests.exceptions import HTTPError
import requests

from .types import AnySNSPayload, SubscriptionConfirmation

logger = logging.getLogger(__name__)

NOTIFICATION_HASH_FORMAT = """Message
{Message}
MessageId
{MessageId}
Subject
{Subject}
Timestamp
{Timestamp}
TopicArn
{TopicArn}
Type
{Type}
"""

NOTIFICATION_HASH_FORMAT_NO_SUBJECT = """Message
{Message}
MessageId
{MessageId}
Timestamp
{Timestamp}
TopicArn
{TopicArn}
Type
{Type}
"""

SUBSCRIPTION_HASH_FORMAT = """Message
{Message}
MessageId
{MessageId}
SubscribeURL
{SubscribeURL}
Timestamp
{Timestamp}
Token
{Token}
TopicArn
{TopicArn}
Type
{Type}
"""


def confirm_subscription(payload: SubscriptionConfirmation) -> HttpResponse:
    """
    Confirm subscription request by making a
    get request to the required url.
    """
    pattern = getattr(
        settings,
        "SNS_SUBSCRIBE_DOMAIN_REGEX",
        r"sns.[a-z0-9\-]+.amazonaws.com$",
    )
    if not payload.SubscribeURL.host or not re.search(
        pattern, payload.SubscribeURL.host
    ):
        logger.error("Invalid Subscription Domain %s", payload.SubscribeURL)
        return HttpResponseBadRequest("Improper Subscription Domain")

    try:
        response = requests.get(str(payload.SubscribeURL))
        response.raise_for_status()
    except HTTPError as e:
        logger.error(
            "HTTP verification Error",
            extra={
                "error": e,
                "sns_payload": payload,
            },
        )
        raise e

    return HttpResponse("OK")


def verify_notification(payload: AnySNSPayload) -> bool:
    """
    Verify notification came from a trusted source
    Returns True if verified, False if not
    """
    cert = get_x509_cert(str(payload.SigningCertURL))
    public_key = cert.public_key()
    if payload.Type == "Notification":
        if payload.Subject is not None:
            hash_format = NOTIFICATION_HASH_FORMAT
        else:
            hash_format = NOTIFICATION_HASH_FORMAT_NO_SUBJECT
    else:
        hash_format = SUBSCRIPTION_HASH_FORMAT
    message = hash_format.format(**payload.model_dump()).encode("utf-8")

    hash_type: hashes.SHA1 | hashes.SHA256
    if payload.SignatureVersion == "1":
        hash_type = hashes.SHA1()
    elif payload.SignatureVersion == "2":
        hash_type = hashes.SHA256()
    else:
        raise ValueError("Unknown SignatureVersion: %s" % payload.SignatureVersion)
    pss = padding.PKCS1v15()
    try:
        if not isinstance(public_key, RSAPublicKey):
            raise ValueError("Unknown key type: %s" % public_key)
        public_key.verify(payload.Signature, message, pss, hash_type)
    except InvalidSignature as e:
        logger.error("Verification of signature raised an Error: %s", e)
        return False
    return True


@lru_cache(maxsize=128)
def get_x509_cert(cert_url: str) -> x509.Certificate:
    """
    Acquire the keyfile
    SNS keys expire and Amazon does not promise they will use the same key
    for all SNS requests. So we need to keep a copy of the cert in our
    cache
    """
    try:
        response = requests.get(cert_url)
        response.raise_for_status()
    except HTTPError as e:
        logger.error("Unable to fetch the keyfile: %s" % e)
        raise
    pemfile = smart_bytes(response.text)
    # Extract the first certificate in the file and confirm it's a valid
    # PEM certificate
    cert = x509.load_pem_x509_certificate(pemfile)
    return cert
